(*Andrew Willouer
COMS 4115 Programming Languages and Translators
           8/10/2009*)

(*Interpret.ml
Takes the abstract syntax tree and runs the program*)

open Ast

module NameMap = Map.Make(struct
  type t = string 
  let compare x y = Pervasives.compare x y
end)

exception ReturnException of Ast.expr * Ast.expr NameMap.t

(*Global variable, that contains all the cards*)
let mydeckstringref = ref ["2S", 2; "3S", 3; "4S", 4; "5S", 5; "6S", 6; "7S", 7; "8S", 8; "9S", 9; "10S", 10; "JS", 10; "QS", 10; "KS", 10; "AS", 10; 
                  "2C", 2; "3C", 3; "4C", 4; "5C", 5; "6C", 6; "7C", 7; "8C", 8; "9C", 9; "10C", 10; "JC", 10; "QC", 10; "KC", 10; "AC", 10;
                  "2D", 2; "3D", 3; "4D", 4; "5D", 5; "6D", 6; "7D", 7; "8D", 8; "9D", 9; "10D", 10; "JD", 10; "QD", 10; "KD", 10; "AD", 10;
                          "2H", 2; "3H", 3; "4H", 4; "5H", 5; "6H", 6; "7H", 7; "8H", 8; "9H", 9; "10H", 10; "JH", 10; "QH", 10; "KH", 10; "AH", 10;];;
                          
(*For future development*)
let deck = []  
let players = []
let cards = []  

(* Main entry point: run a program *)
let run (vars, funcs) =
  (* Put function declarations in a symbol table *)
  let func_decls = List.fold_left
      (fun funcs fdecl -> NameMap.add fdecl.fname fdecl funcs)
      NameMap.empty funcs
  in

  (* Invoke a function and return an updated global symbol table *)
  let rec call fdecl actuals globals =

    (* Evaluate an expression and return (value, updated environment) *)
    let rec eval env = function 
        Noexpr -> Noexpr, env (* must be non-zero for the for loop predicate *)
      | Intliteral(i) -> Intliteral(i), env
      | Stringliteral(s) -> Stringliteral(s), env
      | Boolliteral(i) -> Boolliteral(i), env
      | Cardliteral(i) -> Cardliteral(i), env
      | Handliteral(h, i) -> Handliteral(h, i), env
      | Id(var) ->
	  let locals, globals = env in
	     if NameMap.mem var locals then
	       (NameMap.find var locals), env
	     else if NameMap.mem var globals then
	       (NameMap.find var globals), env
        else raise (Failure ("undeclared identifier " ^ var))
        | Binop(e1, op, e2) ->
	     let v1, env = eval env e1 in
         let v2, env = eval env e2 in
	     let boolean i = if i then Boolliteral(true) else Boolliteral(false) in
	        (match v1, op, v2 with
                                  (*Operators*)
        	    Intliteral(i1), Add, Intliteral(i2) -> Intliteral(i1 + i2)
              | Stringliteral(s1), Add, Intliteral(i2) -> Intliteral(int_of_string s1 + i2)
              | Intliteral(i1), Add, Stringliteral(s2) -> Intliteral(i1 + int_of_string s2)
              | Stringliteral(s1), Add, Stringliteral(s2) -> Intliteral(int_of_string s1 + int_of_string s2)
              | Intliteral(i1), Sub, Intliteral(i2) -> Intliteral(i1 - i2)
              | Stringliteral(s1), Sub, Intliteral(i2) -> Intliteral(int_of_string s1 - i2)
              | Intliteral(i1), Sub, Stringliteral(s2) -> Intliteral(i1 - int_of_string s2)
              | Stringliteral(s1), Sub, Stringliteral(s2) -> Intliteral(int_of_string s1 - int_of_string s2)
              | Intliteral(i1), Mult, Intliteral(i2) -> Intliteral(i1 * i2)
              | Stringliteral(s1), Mult, Intliteral(i2) -> Intliteral(int_of_string s1 * i2)
              | Intliteral(i1), Mult, Stringliteral(s2) -> Intliteral(i1 * int_of_string s2)
              | Stringliteral(s1), Mult, Stringliteral(s2) -> Intliteral(int_of_string s1 * int_of_string s2)
              | Intliteral(i1), Div, Intliteral(i2) -> Intliteral(i1 / i2)
              | Stringliteral(s1), Div, Intliteral(i2) -> Intliteral(int_of_string s1 / i2)
              | Intliteral(i1), Div, Stringliteral(s2) -> Intliteral(i1 / int_of_string s2)
              | Stringliteral(s1), Div, Stringliteral(s2) -> Intliteral(int_of_string s1 / int_of_string s2)
              | Handliteral(a1, b1), Equal, Intliteral(i2) -> boolean (b1 = i2)
              | Stringliteral(s1), Equal, Intliteral(i2) -> boolean (int_of_string s1  = i2)  
              | Intliteral(i1), Equal, Intliteral(i2) -> boolean (i1 = i2)
              | Stringliteral(s1), Equal, Stringliteral(s2) -> boolean (s1  = s2)  
              | Cardliteral(c1), Equal, Cardliteral(c2) -> boolean (c1 = c2)
              | Boolliteral(b1), Equal, Boolliteral(b2) -> boolean (string_of_bool b1 = string_of_bool b2)
              | Intliteral(i1), Neq, Intliteral(i2) -> boolean(i1 != i2) 
              | Stringliteral(s1), Neq, Stringliteral(s2) -> boolean(s1 != s2)
              | Stringliteral(s1), Neq, Intliteral(i2) -> boolean (int_of_string s1  != i2)
              | Cardliteral(c1), Neq, Cardliteral(c2) -> boolean(c1 != c2) 
              | Boolliteral(b1), Neq, Boolliteral(b2) -> boolean(string_of_bool b1 != string_of_bool b2) 
              | Intliteral(i1), Less, Intliteral(i2) -> boolean(i1 < i2) 
              | Stringliteral(s1), Less, Stringliteral(s2) -> boolean(s1 < s2) 
              | Cardliteral(c1), Less, Cardliteral(c2) -> boolean(c1 < c2)
              | Stringliteral(s1), Less, Intliteral(i2) -> boolean(int_of_string s1 < i2)
              | Intliteral(i1), Leq, Intliteral(i2) -> boolean(i1 <= i2) 
              | Stringliteral(s1), Leq, Stringliteral(s2) -> boolean(s1 <= s2)
              | Cardliteral(c1), Leq, Cardliteral(c2) -> boolean(c1 <= c2)
              | Stringliteral(s1), Leq, Intliteral(i2) -> boolean(int_of_string s1 <= i2)
              | Handliteral(a1, b1), Greater, Handliteral(a2, b2) -> boolean(b1 > b2)
              | Intliteral(i1), Greater, Intliteral(i2) -> boolean(i1 > i2) 
              | Stringliteral(s1), Greater, Stringliteral(s2) -> boolean(s1 > s2) 
              | Cardliteral(c1), Greater, Cardliteral(c2) -> boolean(c1 > c2)  
              | Stringliteral(s1), Greater, Intliteral(i2) -> boolean(int_of_string s1 > i2)
              | Intliteral(i1), Geq, Intliteral(i2) -> boolean(i1 >= i2) 
              | Stringliteral(s1), Geq, Stringliteral(s2) -> boolean(s1 >= s2) 
              | Cardliteral(c1), Geq, Cardliteral(c2) -> boolean(c1 >= c2)
              | Stringliteral(s1), Geq, Intliteral(i2) -> boolean(int_of_string s1 >= i2)
              | Boolliteral(b1), And, Boolliteral(b2) -> boolean(b1 && b2)
              | Boolliteral(b1), Or, Boolliteral(b2) -> boolean(b1 || b2)
              | _, _, _ -> raise (Failure ("Invalid binary operation"))
             ) , env
              | Variable(var) ->
                let locals, globals = env in
                (match var with
                  VarExp(id, scope) ->
                    (match scope with
                      Local ->
                       if NameMap.mem id locals then
                       NameMap.find id locals, env
                       else raise (Failure ("undeclared local identifier " ^ id))
                     | Global -> 
                       if NameMap.mem id globals then
                       NameMap.find id globals, env
                       else raise (Failure ("undeclared global identifier " ^ id))
                     )
                 )
(*Assign finds the variable in the map and then binds(assigns) the variable*)
      | Assign(var, e) ->
    	  let v, (locals, globals) = eval env e in
        (match var with
          VarExp(id, scope) ->
            (match scope with
              Local ->
                if NameMap.mem id locals then
                  v, (NameMap.add id v locals, globals)
                else raise (Failure ("undeclared local identifier " ^ id))
              | Global -> 
                if NameMap.mem id globals then
                  v, (locals, NameMap.add id v globals)
                else raise (Failure ("undeclared global identifier " ^ id))
             )
        )
     | Call(f, actuals) ->
	  let fdecl =
	    try NameMap.find f func_decls
	    with Not_found -> raise (Failure ("undefined function " ^ f))
	  in
	  let actuals, env = List.fold_left
	      (fun (actuals, values) actual ->
		let v, env = eval env actual in 
         List.append actuals [v], values) ([], env) actuals
	  in
	  let (locals, globals) = env in
	  try
	    let globals = call fdecl actuals globals 
     in Boolliteral(false), (locals, globals)
     with ReturnException(v, globals) -> v, (locals, globals)
     
    in

    (* Execute a statement and return an updated environment *)
    let rec exec env = function
       Expr(e) -> let _, env = eval env e in env
                                 (*If expressions*)
      | If(e, s1, s2) ->
	  let v, env = eval env e in
        let b = (match v with
           Boolliteral(b) -> b
           | _ -> raise (Failure("Invalid conditional expression.")))
             in
              if b then List.fold_left exec env (List.rev s1)
              else List.fold_left exec env (List.rev s2)
                                 (*While expressions*)
      | While(e, s) ->
	  let rec loop env =
	    let v, env = eval env e in
          let b = (match v with
            Boolliteral(b) -> b
            | _ -> raise (Failure("Invalid conditional expression.")))
             in
              if b then loop (List.fold_left exec env (List.rev s))
              else env 
                in loop env
                             (*For expression*)
     | For(e1, e2, e3, s) ->
	  let _, env = eval env e1 in
	  let rec loop env =
	    let v, env = eval env e2 in
          let b = (match v with
            Boolliteral(b) -> b
            | _ -> raise (Failure("Invalid conditional expression.")))
             in
              if b then let _, env = eval (exec env s) e3 in
	      loop env
	    else
	      env
         in loop env
    (*Read takes input from the user and assigns it to a variable*)
      | Read(var) ->
        let input = read_line() in
          let v = (match input with
             a -> Stringliteral(a)
      |_ -> raise(Failure("Invalid input"))) in
          let ret, env = eval env (Assign(var, v)) in env
   (*Print, prints out a variable*)
    | Print(e) ->
          let v, env = eval env e in
          begin
            let str = (match v with
              Boolliteral(b) -> string_of_bool b
            | Intliteral(i) -> string_of_int i
            | Cardliteral(c) -> "[Card: " ^ c ^ "]"
            | Stringliteral(s) -> s
            | Handliteral(a, b) -> a ^ " value is " ^ string_of_int b
            | _ -> raise (Failure ("Invalid print expression.")))
          in
          print_endline str;
          env
         end
     (*Deal, takes the topcard from the mydeckstringref, removes that card from the deckstring
              and assigns the  card to a variable*)
     | Deal(var) -> let topcard = List.hd !mydeckstringref  in
      mydeckstringref := List.tl !mydeckstringref;        
     (*Remove topcard from the deck*)
     let v = (match topcard with
             a,b -> Handliteral(a,b)
      |_ -> raise(Failure("Invalid input"))) in
     let ret, env = eval env (Assign(var, v)) in env
      (*Value prints the int part of a handliteral, the card's value*)  
     | Value(e) -> let v, env = eval env e in
          begin
            let str = (match v with
            Handliteral(a, b) ->string_of_int b
            | _ -> raise (Failure ("Invalid Value expression.")))
          in
          print_endline str;
          env
         end        
   | Return(e) ->
	  let v, (locals, globals) = eval env e in
	  raise (ReturnException(v, globals))
    in

    (* Enter the function: bind actual values to formal arguments *)
    let locals =
      try List.fold_left2
	  (fun locals formal actual -> NameMap.add formal actual locals)
	  NameMap.empty fdecl.formals actuals
      with Invalid_argument(_) ->
	raise (Failure ("wrong number of arguments passed to " ^ fdecl.fname))
    in
    (* Initialize local variables to 0 *)
    let locals = List.fold_left
	(fun locals local -> NameMap.add local (Intliteral(0)) locals) locals fdecl.locals
    in
    (* Execute each statement in sequence, return updated global symbol table *)
    snd (List.fold_left exec (locals, globals) fdecl.body)

  (* Run a program: initialize global variables to 0, find and run "main" *)
  in let globals = List.fold_left
      (fun globals vdecl -> NameMap.add vdecl (Intliteral(0)) globals) NameMap.empty vars
           
  in try
    call (NameMap.find "main" func_decls) [] globals
  with Not_found -> raise (Failure ("did not find the main() function"))
