(*Andrew Willouer
COMS 4115 Programming Languages and Translators
8/10/2009*)

(*Printer.ml
Used to print out the AST*)

open Ast

let rec tabs i = match i with
  0 -> ""
  | x -> "\t" ^ tabs (x - 1)

let string_of_op op = match op with
    Add -> "+"
  | Sub -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Equal -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq -> ">="
  | And -> "&&"
  | Or -> "||"

let rec string_of_types types = match types with
    Int -> "int"
  | StringType -> "string"
  | Bool -> "bool"
  | Card -> "Card"
  | Deck -> "Deck" 
  | Hand -> "Hand"
  | Player -> "Player"
    
let string_of_scope scope = match scope with
  Global ->"@"
  | Local -> ""

let rec string_of_varexp v = match v with
  | VarExp(id, s) -> string_of_scope s ^ id
and string_of_expr expr = match expr with (*function*)
    Intliteral(l) -> string_of_int l
  | Variable(v) -> string_of_varexp v 
  | Boolliteral(b) -> string_of_bool b
  | Cardliteral(c) -> c
  | Stringliteral(s) -> "\"" ^ s ^ "\""
  | Id(s) -> s
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^
      (match o with
	Add -> "+" | Sub -> "-" | Mult -> "*" | Div -> "/"
      | Equal -> "==" | Neq -> "!="
                      | Less -> "<" | Leq -> "<=" | Greater -> ">" | Geq -> ">=" | And -> "&&" | Or -> "||") ^ " " ^
      string_of_expr e2
  | Assign(v, e) -> string_of_varexp v ^ " = " ^ string_of_expr e
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Noexpr -> ""
    
let string_of_vardec v = match v with
  id -> "var" ^ id

let rec string_of_stmt t stmt = tabs t ^ match stmt with (*function*)
  (*Block(stmts) ->
  "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n")*)
   Expr(expr) -> string_of_expr expr ^ ";\n";
  | Print(expr) -> "Print " ^ string_of_expr expr ^ ";\n";
  | Read(var) -> "Read " ^ string_of_varexp var ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      String.concat "" (List.map (string_of_stmt (t+1)) s1) ^
        tabs t ^ "} else {\n" ^
      String.concat "" (List.map (string_of_stmt (t+1)) s2) ^
        tabs t ^ "} else {\n"    
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") {\n" ^
    String.concat "" (List.map (string_of_stmt (t+1)) s) ^
        tabs t ^ "} \n"

let string_of_vdecl id = "int " ^ id ^ ";\n"

let string_of_fdecl fdecl =
  fdecl.fname ^ "(" ^ String.concat ", " fdecl.formals ^ ")\n{\n" ^
  String.concat "" (List.map string_of_vdecl fdecl.locals) ^
  String.concat "" (List.map (string_of_stmt 1) fdecl.body) ^
  "}\n"

let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
 
  
let string_of_sdecl_g sname vars =
  sname ^ "\n{\n" ^
  String.concat "" (List.map string_of_vdecl vars) ^ "}\n"
  
let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
  
(*let string_of_program (spec, funcs) =
  string_of_sdecl_g "Globals" spec.glob.globals ^ "\n" ^
String.concat "\n" (List.map string_of_fdecl funcs)*)
  
 
